#include "bmp.hpp"
#include "crc32.hpp"
using namespace std;

struct BitmapFileHeader
{
	char Magic[2];
	unsigned int FileSize;
	unsigned int tmp;
	unsigned int ImageOffset;
} __attribute__ ((packed));

struct BitmapInfoHeader
{
	unsigned int InfoHeaderSize;
	int Width;
	int Height;
	unsigned short Planes;
	unsigned short BitSize;
	unsigned int Compress;
	unsigned int HDpi;
	unsigned int VDpi;
	unsigned int PalUsed;
	unsigned int PalImportant;
} __attribute__ ((packed));

bool 
Bitmap::load(string filename)
{
	int now_size = bmp.size();
	if (isLoaded(filename)) {
		return true;
	}

	FILE *file;
	if ((file = fopen(filename.c_str(), "rb")) == NULL) {
		fprintf(stderr, "Can't open file %s\n", filename.c_str());
		return false;
	}
	fseek(file, 0, SEEK_END);
	int size = ftell(file);
	fseek(file, 0, SEEK_SET);

	char* buf = new char[size];
	fread(buf, 1, size, file);
	fclose(file);

	bmp.push_back(buf);
	if (now_size != bmp.size()) {
		name.insert(pair<string, int>(string(filename), now_size));
	}
	return true;
}

int 
Bitmap::getIndex(string filename)
{
	map<string, int>::iterator it = name.find(filename);
	if (it == name.end())
		return -1;
	else
		return it->second;
}

int 
Bitmap::getWidth(string filename)
{
	int index = getIndex(filename);
	if (index < 0) return 0;
	BitmapFileHeader *bh = (BitmapFileHeader*)(bmp[index]);
	BitmapInfoHeader *bi = (BitmapInfoHeader*)(&bh[1]);
	
	return bi->Width;
}

int 
Bitmap::getHeight(string filename)
{
	int index = getIndex(filename);
	if (index < 0) return 0;
	BitmapFileHeader *bh = (BitmapFileHeader*)(bmp[index]);
	BitmapInfoHeader *bi = (BitmapInfoHeader*)(&bh[1]);
	
	return bi->Height;
}

static unsigned short 
rev_nibble_word(unsigned short s)
{
	return ((s & 0xf000) >> 12)
			| ((s & 0x0f00) >> 4)
			| ((s & 0x00f0) << 4)
			| ((s & 0x000f) << 12);
}

int 
Bitmap::getIndexSize(void)
{
	return spr.size();
}

unsigned short* 
Bitmap::getIndexedSprPointer(int index)
{
	if (index >= spr.size())
		index = spr.size() - 1;

	return &spr[index].sp[0][0];
}

int 
Bitmap::generate16x16(string filename, int x, int y, int *hv_) {
	int index = getIndex(filename);
	if (index < 0) return -1;

	sp_ sp;

	for (int yy = 0; yy < 2; yy++) {
		for (int xx = 0; xx < 2; ++xx) {
			unsigned short* buf = generate8x8(filename, x + xx*8, y + yy*8);

			int vx = xx;
			int vy = 1 - yy;
			int hx = 1 - xx;
			int hy = yy;
			for (int i = 0; i < 16; i+=2) {
				sp.sp[NORM  ][yy * 32 + xx * 16 + i  ] = buf[i];
				sp.sp[NORM  ][yy * 32 + xx * 16 + i+1] = buf[i+1];

				sp.sp[HFLIP ][hy * 32 + hx * 16 + i  ] = rev_nibble_word(buf[i+1]);
				sp.sp[HFLIP ][hy * 32 + hx * 16 + i+1] = rev_nibble_word(buf[i]);

				sp.sp[VFLIP ][vy * 32 + vx * 16 + i  ] = buf[14-i];
				sp.sp[VFLIP ][vy * 32 + vx * 16 + i+1] = buf[14-i+1];

				sp.sp[HVFLIP][vy * 32 + hx * 16 + i  ] = rev_nibble_word(buf[14-i+1]);
				sp.sp[HVFLIP][vy * 32 + hx * 16 + i+1] = rev_nibble_word(buf[14-i]);
			}
			delete [] buf;
		}
	}
	int blank = 1;
	for (int i = 0; i < 64; i++) {
		if (sp.sp[0][i] != 0) {
			blank = 0;
			break;
		}
	}
	if (blank) return -1;

	for (int i = 0; i < 4; i++) {
		sp.crc[i] = crc32((char*)&sp.sp[i][0], 128);
	}

	static const int hv_table[4][4] = {
		{  NORM,  HFLIP,  VFLIP, HVFLIP, }, // NORM -> N,H,V,HV
		{ HFLIP,   NORM, HVFLIP,  VFLIP, }, // HFLIP -> N,H,V,HV
		{ VFLIP, HVFLIP,   NORM,  HFLIP, }, // VFLIP -> N,H,V,HV
		{HVFLIP,  VFLIP,  HFLIP,   NORM, }, // HVFLIP -> N,H,V,HV
	};

	bool hit = false;
	int num = 0;
	int hv = 0;
	for (int i = 0; i < spr.size(); i++) {
		for (int j = 0; j < 4; j++) {
			if (sp.crc[NORM  ] == spr[i].crc[j]) { num = i; hit = true; hv = hv_table[NORM][j]; break; }
			if (sp.crc[HFLIP ] == spr[i].crc[j]) { num = i; hit = true; hv = hv_table[HFLIP][j]; break; }
			if (sp.crc[VFLIP ] == spr[i].crc[j]) { num = i; hit = true; hv = hv_table[VFLIP][j]; break; }
			if (sp.crc[HVFLIP] == spr[i].crc[j]) { num = i; hit = true; hv = hv_table[HVFLIP][j]; break; }
		}
		if (hit) break;
	}
	if (!hit) {
		spr.push_back(sp);
		num = spr.size() -1;
	} else {
		*hv_ = hv_table[*hv_][hv];
	}

	return num;
}

unsigned short* 
Bitmap::generate8x8(string filename, int x, int y) {
	int index = getIndex(filename);

	BitmapFileHeader *bh = (BitmapFileHeader*)(bmp[index]);
	BitmapInfoHeader *bi = (BitmapInfoHeader*)(&bh[1]);

	char* b = bmp[index] + bh->ImageOffset;
	unsigned short *buf = new unsigned short[16];

	for (int yy = 0; yy < 8; yy++) {
		char* bx = &b[(bi->Height - (y + yy) - 1) * bi->Width + x];
		for (int xx = 0; xx <= 1; xx++) {
			buf[yy * 2 + xx] = ((bx[3]&15) << 12)
							 | ((bx[2]&15) << 8)
							 | ((bx[1]&15) << 4)
							 |  (bx[0]&15);
			bx +=4;
		}
	}
	return buf;
}
